"""
Spaces defined for the algorithms to user
It might be a good idea to experiment with some of the distributions here
"""

from math import log
from hyperopt import hp
from hyperopt.pyll.base import scope

LR_LOG_LOWER = log(0.0001)
LR_LOG_UPPER = log(0.1)

# NOTE: Do not use keyword arguments for hyperopt distributions. It causes a crash when using ATPE

SVD_DEFAULT_SPACE = {
    'n_factors': scope.int(hp.quniform('n_factors', 1, 200, 1)),
    'n_epochs': scope.int(hp.quniform('n_epochs', 1, 200, 1)),
    'lr_bu': hp.loguniform('lr_bu', LR_LOG_LOWER, LR_LOG_UPPER),
    'lr_bi': hp.loguniform('lr_bi', LR_LOG_LOWER, LR_LOG_UPPER),
    'lr_pu': hp.loguniform('lr_pu', LR_LOG_LOWER, LR_LOG_UPPER),
    'lr_qi': hp.loguniform('lr_qi', LR_LOG_LOWER, LR_LOG_UPPER),
    'reg_bu': hp.loguniform('reg_bu', LR_LOG_LOWER, LR_LOG_UPPER),
    'reg_bi': hp.loguniform('reg_bi', LR_LOG_LOWER, LR_LOG_UPPER),
    'reg_pu': hp.loguniform('reg_pu', LR_LOG_LOWER, LR_LOG_UPPER),
    'reg_qi': hp.loguniform('reg_qi', LR_LOG_LOWER, LR_LOG_UPPER)
}

SVDPP_SPACE = {
    **SVD_DEFAULT_SPACE,
    'lr_yj': hp.loguniform('lr_yj', LR_LOG_LOWER, LR_LOG_UPPER),
    'reg_yj': hp.loguniform('reg_yj', LR_LOG_LOWER, LR_LOG_UPPER)
}

# TODO: Shrinkage param is only used when pearson_baseline is selected. It would
# be great if we could conditionally optimize if pearson_baseline is used
SIMILARITY_OPTIONS_SPACE = {
    'name': hp.choice('name', ['cosine', 'msd', 'pearson', 'pearson_baseline']),
    'user_based': hp.choice('user_based', [False, True]),
    'min_support': scope.int(hp.quniform('min_support', 1, 100, 1)),
    'shrinkage': scope.int(hp.quniform('shrinkage', 1, 300, 1))
}

BSL_OPTIONS_SPACE = hp.choice('bsl_options', [
    {
        'method': 'als',
        'reg_i': hp.uniform('reg_i', 1, 100),
        'reg_u': hp.uniform('reg_u', 1, 100),
        'n_epochs': scope.int(hp.quniform('n_epochs', 5, 200, 1)),
    },
    {
        'method': 'sgd',
        'reg': hp.loguniform('reg', LR_LOG_LOWER, LR_LOG_UPPER),
        'learning_rate': hp.loguniform('learning_rate', LR_LOG_LOWER, LR_LOG_UPPER)
    }
])

BASELINE_ONLY_SPACE = {
    'bsl_options': BSL_OPTIONS_SPACE
}

KNN_DEFAULT_SPACE = {
    'k': scope.int(hp.quniform('k', 1, 500, 1)),
    'min_k': scope.int(hp.quniform('min_k', 1, 50, 1)),
    'sim_options': SIMILARITY_OPTIONS_SPACE
}

KNN_BASELINE_SPACE = {
    **KNN_DEFAULT_SPACE,
    'bsl_options': BSL_OPTIONS_SPACE
}

NMF_DEFAULT_SPACE = {
    'n_factors': scope.int(hp.quniform('n_factors', 1, 500, 1)),
    'n_epochs': scope.int(hp.quniform('n_epochs', 5, 200, 1)),
    'lr_bu': hp.loguniform('lr_bu', LR_LOG_LOWER, LR_LOG_UPPER),
    'lr_bi': hp.loguniform('lr_bi', LR_LOG_LOWER, LR_LOG_UPPER),
    'reg_bu': hp.loguniform('reg_bu', LR_LOG_LOWER, LR_LOG_UPPER),
    'reg_bi': hp.loguniform('reg_bi', LR_LOG_LOWER, LR_LOG_UPPER),
    'reg_pu': hp.loguniform('reg_pu', LR_LOG_LOWER, LR_LOG_UPPER),
    'reg_qi': hp.loguniform('reg_qi', LR_LOG_LOWER, LR_LOG_UPPER),
    'biased': hp.choice('biased', [False, True])
}

CO_CLUSTERING_DEFAULT_SPACE = {
    'n_cltr_u': scope.int(hp.quniform('n_cltr_u', 1, 1000, 1)),
    'n_cltr_i': scope.int(hp.quniform('n_cltr_i', 1, 100, 1)),
    'n_epochs': scope.int(hp.quniform('n_epochs', 5, 200, 1)),
}



import logging
import multiprocessing
from auto_surprise.constants import (EVALS_MULTIPLIER, MAX_WORKERS)
from auto_surprise.trainer import Trainer
from auto_surprise.strategies.base import StrategyBase

class ContinuousParallel(StrategyBase):
    """
    Executes all alogrithms in parallel until time limit exceeded or max evals reached.
    """
    def __init__(self, *args):
        self.__logger = logging.getLogger(__name__)
        super().__init__(*args)

    def evaluate(self):
        """
        Evaluate performance of algorithms
        """

        self.__logger.info("Starting evaluation using strategy : ContinuousParallel")

        max_evals = self.max_evals
        processes = []

        # Run for N algorithms
        with multiprocessing.Manager() as mp_manager:

            tasks = mp_manager.dict()

            for algo in self.algorithms:
                self.__logger.debug("Starting process with %s algorithm" % algo)

                trainer = Trainer(self.tmp_dir, algo=algo, data=self.data, target_metric=self.target_metric, hpo_algo=self.hpo_algo, debug=self._debug)
                p = multiprocessing.Process(target=trainer.start_with_limits, args=(max_evals, self.time_limit, tasks))
                processes.append(p)
                p.start()

            # Wait for processes to complete
            for process in processes:
                process.join()

            best_model = min(tasks.items(), key=(lambda x: x[1]['score']['loss']))[0]
            best_params = tasks[best_model]['score']['hyperparams']
            best_score = tasks[best_model]['score']['loss']

            return best_model, best_params, best_score, tasks.copy()



from auto_surprise.constants import AVAILABLE_METRICS
from auto_surprise.exceptions import ValidationError
from surprise.dataset import DatasetAutoFolds

def validate_target_metric(metric):
    if metric not in AVAILABLE_METRICS:
        raise ValidationError('target_metric', 'Unknown target metric : "%s". Must be in %s' % (metric, AVAILABLE_METRICS))
    else:
        return True

def validate_dataset(data):
    """
    Validate user given dataset
    """
    if data:
        if not isinstance(data, DatasetAutoFolds):
            raise ValidationError('data', 'Unknown data format. Must be and instance of "DatasetAutoFolds". Got "%s"' % type(data))
        else:
            return True
    else:
        raise ValidationError('data', 'Data must be present. Got: %s' % data)

def validate_max_evals(max_evals):
    """
    Max evaluations for hyperparameter tuning. Must be an unsigned integer greater than 0
    """
    if not isinstance(max_evals, int) or not max_evals > 0:
        raise ValidationError('max_evals', 'max_evals must be an unsigned integer greater than 0. Got {0}'.format(max_evals))
    else:
        return True



import os
import logging
import pathlib

from auto_surprise.constants import (
    DEFAULT_TARGET_METRIC, DEFAULT_MAX_EVALS,
    FULL_ALGO_LIST, QUICK_COMPUTE_ALGO_LIST,
    BASELINE_ALGO, EVALS_MULTIPLIER,
    SURPRISE_ALGORITHM_MAP, DEFAULT_HPO_ALGO
)
from auto_surprise.trainer import Trainer
from auto_surprise.exceptions import ValidationError
from auto_surprise.context.backend import BackendContextManager
from auto_surprise.strategies.basic_reduction import BasicReduction
from auto_surprise.strategies.continuous_parallel import ContinuousParallel
import auto_surprise.validation_util as validation_util

class Engine(object):
    def __init__(self, debug=False):
        """
        Initialize new engine
        """
        logging.basicConfig(level=logging.INFO)
        self.__logger = logging.getLogger(__name__)

        self._debug = debug
        self._current_path = pathlib.Path().absolute()

    def train(
        self,
        target_metric=DEFAULT_TARGET_METRIC,
        data=None,
        max_evals=DEFAULT_MAX_EVALS,
        quick_compute=False,
        cpu_time_limit=None,
        hpo_algo=DEFAULT_HPO_ALGO,
        strategy="continuos_parallel"
    ):
        """
        Train and find most optimal model and hyperparameters
        """

        try:
            # Validations
            validation_util.validate_target_metric(target_metric)
            validation_util.validate_dataset(data)
            validation_util.validate_max_evals(max_evals)
        except ValidationError as err:
            """
            Catch validation errors
            """
            logging.critical(err.message)
            raise err

        # Determine baseline value from random normal predictor
        with BackendContextManager(self._current_path) as tmp_dir:
            self.__logger.info("Available CPUs: {0}".format(os.cpu_count()))
            baseline_trainer = Trainer(tmp_dir, algo=BASELINE_ALGO, data=data, target_metric=target_metric, debug=self._debug)
            baseline_loss = baseline_trainer.start(1)[1]['loss']
            self.__logger.info("Baseline loss : {0}".format(baseline_loss))

            algorithms = QUICK_COMPUTE_ALGO_LIST if quick_compute else FULL_ALGO_LIST

            # Select the strategy
            if strategy == 'continuos_parallel':
                strategy = ContinuousParallel(
                    algorithms,
                    data,
                    target_metric,
                    baseline_loss,
                    tmp_dir,
                    max_evals,
                    cpu_time_limit,
                    hpo_algo,
                    self._debug
                )
            else:
                strategy = BasicReduction(
                    algorithms,
                    data,
                    target_metric,
                    baseline_loss,
                    tmp_dir,
                    hpo_algo=hpo_algo,
                    debug=self._debug
                )

            best_model, best_params, best_score, tasks = strategy.evaluate()

        return best_model, best_params, best_score, tasks

    def build_model(self, algo_name, params):
        algo = SURPRISE_ALGORITHM_MAP[algo_name]
        if (params):
            return algo(**params)
        else:
            return algo()



import sys
import traceback
import logging
from surprise.model_selection import cross_validate

# import algorithms
from auto_surprise.algorithms.auto_surprise_svd import AutoSurpriseSVD
from auto_surprise.algorithms.auto_surprise_svdpp import AutoSurpriseSVDpp
from auto_surprise.algorithms.auto_surprise_nmf import AutoSurpriseNMF
from auto_surprise.algorithms.auto_surprise_knn_baseline import AutoSurpriseKNNBaseline
from auto_surprise.algorithms.auto_surprise_knn_basic import AutoSurpriseKNNBasic
from auto_surprise.algorithms.auto_surprise_knn_with_means import AutoSurpriseKNNWithMeans
from auto_surprise.algorithms.auto_surprise_knn_with_z_score import AutoSurpriseKNNWithZScore
from auto_surprise.algorithms.auto_surprise_slope_one import AutoSurpriseSlopeOne
from auto_surprise.algorithms.auto_surprise_co_clustering import AutoSurpriseCoClustering
from auto_surprise.algorithms.auto_surprise_normal_predictor import AutoSurpriseNormalPredictor
from auto_surprise.algorithms.auto_surprise_baseline_only import AutoSurpriseBaselineOnly

from auto_surprise.constants import ALGORITHM_MAP, DEFAULT_TARGET_METRIC, CV_N_JOBS, DEFAULT_HPO_ALGO
from auto_surprise.context.limits import TimeoutException
from auto_surprise.context.result_logging_manager import ResultLoggingManager
import auto_surprise.context.limits as limits

class Trainer(object):
    def __init__(self, tmp_dir, algo=None, data=None, target_metric=DEFAULT_TARGET_METRIC, hpo_algo=DEFAULT_HPO_ALGO, debug=False):
        """
        Initialize new trainer
        """
        self.__logger = logging.getLogger(__name__)
        self._debug = debug
        self._tmp_dir = tmp_dir
        self._algo_name = algo
        self._algo_class = ALGORITHM_MAP[algo]
        # Dynamically instantiate algorithm
        self.algo = getattr(sys.modules[__name__], self._algo_class)(data=data, metric=target_metric, cv_n_jobs=CV_N_JOBS, hpo_algo=hpo_algo, debug=debug)

    def start(self, max_evals):

        try:
            with ResultLoggingManager(self._tmp_dir, self._algo_class) as result_logger:
                self.algo.set_result_logger(result_logger)

                best, trials = self.algo.best_hyperparams(max_evals)

            best_trial = None
            # Sort best trial based on loss value
            if best:
                best_trial = sorted(trials.results, key=lambda x: x['loss'], reverse=False)[0]
            else:
                best_trial = trials

            return best, best_trial
        except Exception as e:
            self.__logger.ERROR('Exception : ', e)
            self.__logger.ERROR(traceback.format_exc())
            if self._debug:
                raise

            return False, False

    def start_with_limits(self, max_evals, time_limit, tasks):
        # Result logger should not be destroyed by exceptions
        with ResultLoggingManager(self._tmp_dir, self._algo_class) as result_logger:
            try:
                # Although hyperopt.fmin does have a timout parameter, it doesnt seem to work with multiprocessing
                with limits.run_with_enforced_limits(time_limit=time_limit):
                    self.algo.set_result_logger(result_logger)

                    best, best_trial = self.algo.best_hyperparams(max_evals)

                    if self.algo.trials.results:
                        best_trial = sorted(self.algo.trials.results, key=lambda x: x['loss'], reverse=False)[0]

                    tasks[self._algo_name] = {
                        'score': best_trial,
                    }

            except TimeoutException:
                # Handle timeout when enforced cpu time limit is reached
                trials = self.algo.trials

                if trials.results:
                    best_trial = sorted(trials.results, key=lambda x: x['loss'], reverse=False)[0]
                    tasks[self._algo_name] = {
                        'score': best_trial,
                    }
                else:
                    best_trial = False
                    tasks[self._algo_name] = {
                        'score': { 'loss': 100, 'hyperparams': None },
                    }

            except Exception:
                self.__logger.error("Exception for algo {0}".format(self._algo_name))
                self.__logger.error(traceback.format_exc())

                if self._debug:
                    raise

                tasks[self._algo_name] = {
                    'score': { 'loss': 100, 'hyperparams': None },
                    'exception': True
                }



from auto_surprise.constants import DEFAULT_TARGET_METRIC, CV_N_JOBS, DEFAULT_HPO_ALGO, DEFAULT_CV_ITERS
from hyperopt import Trials

class AlgorithmBase(object):
    def __init__(self, cv=DEFAULT_CV_ITERS, metric=DEFAULT_TARGET_METRIC, data=None, cv_n_jobs=CV_N_JOBS, hpo_algo=DEFAULT_HPO_ALGO, debug=False):
        self._cv = cv
        self._metric = metric
        self._data = data
        self._debug = debug
        self._cv_n_jobs = cv_n_jobs
        self._hpo_algo = hpo_algo
        self.trials = Trials()

    def set_result_logger(self, result_logger_manager):
        self._result_logger = result_logger_manager
